/*
	aa_main.c - The main module of the Antenna Analyzer (Ant_Anl)
	Copyright (C) 2005 Vadim Demidov vadim_d .at. sourceforge.net

    Ant_Anl is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published
    by the Free Software Foundation; either version 2 of the License,
    or (at your option) any later version.

    Ant_Anl is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Ant_Anl; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	The algorithm of R/X/SWR calculation is reproduced in C from
	PICAXE-28X BASIC software developed by Jim Tregellas VK5JST,
	Keith Gooley VK5OQ, and Lee de Vries VK3PK:
		http://www.users.on.net/~endsodds/newswr_1.bas
	For hardware details see:
		http://www.users.on.net/~endsodds/analsr.htm

	Revision History:

	17 Apr 05 - Created for testing low level modules (VD)
	25 Apr 05 - First functional (?) version, called v0.01 (VD)
	27 Apr 05 - fpcalc(FPC_INITF) added to initialize FP flags
				before starting calculations - this enables
				rounding instead of truncation (VD)
	04 May 05 - Minor comment editing before releasing as v0.02 (VD)
*/

#include "aa_defs.h"	// Definitions common to all Ant_Anl project
#include "glb_defs.h"	// Global functions and data objects

#pragma rambank 0
#pragma codepage 0

/*
	Configuration fuses of PIC16F873A. Currently this feature doesn't
	work in the compiler (version 3.2F), so it is temporarily moved
	to CFG.ASM file
*/

//#pragma config BODEN=off  // Borwn-Out reset off
//#pragma config WDTE=off   // Watchdog off
//#pragma config PWRTE=on   // Power Timer on
//#pragma config FOSC=HS    // Quartz
//#pragma config |=0x2F00   // All other bits

#include <math16.h>

/*
	Messages kept in code EPROM
*/
const char msg_vers[]="VK5JST v0.02";
const char msg_batt[]="BATTERY=";
const char msg_mhzr[]="MHz R";
const char msg_norf[]="NO RF OSCILLATOR";
const char msg_open[]="  OPEN CIRCUIT  ";
const char msg_shrt[]="  SHORT CIRCUIT ";
const char msg_swr[] =" SWR";

/* Local functions */

void lcd_puts(const char * s);	/* write a string of characters to the LCD */
void err_msg(const char * s);	/* Display an error message */
void disp_rx(void);				/* Display R or X */

void main( void )
{
	uns16 i;			// Declare all local variables first
	char *chpt;
	float24 v_in, v_50, v_ld, fp_temp, sq_vin, sq_v50, sq_vld;

	/* Define I/O direction and initial value for each port */

	PORTA = 0b00000000;  // Output 0 to each port before defining I/O
	PORTB = 0b00000000;
//  PORTC = 0b00000000;
	OPTION_REG = 0b00110111;  // Pull up on PORTB, Timer 0 as counter
	TRISA = 0b00101111;  // AN4,AN3-AN0 - inputs, RA4 (T0CKI) - output
	TRISB = 0b00000000;  // All outputs in B
	LED_0 = 1;			// ### diagnostic
	a2d_init();			// Initialize ADC
	lcd_init();			// Initialize the LCD before using it
	fpcalc(FPC_INITF);	// Initialize FP flags - set rounding mode
/*
	Battery test: a voltage on the AN3 input is 1/5 of V_bat, hence
	ADC readings N_bat=(3.9/(16+3.9))*1024*(V_bat/V_ref), V_ref=5V.
	V_bat=N_bat*V_ref*(19.9/3.9)*(1/1024)=N_bat*0.02491486 . To get
	the result in 0.01V units we need V_bat*100=N_bat*2.491486
*/
	lcd_puts(msg_vers);	// Display software version
	lcd_goto(0x40);		// Line 2
	lcd_puts(msg_batt);	// "Battery=" on the line 2
	a2d_read(ACH_VBT);	// Read battery voltage, result in fp_x_u16
	fpcalc(FPC_U16TOFP); // Converted to floating point in fp_x
	fpcalc(FPC_UP);		// Now fp_y is N_bat
	fp_x=2.4915;		// FP constants can be assigned directly
	fpcalc(FPC_MUL);	// V_bat*100 is in fp_x
	fpcalc(FPC_FPTOS24); // Convert to 24-bit binary in fp_x_i24
	u24_bcd32();		// BCD result is in bcd32[]: 33 22 11 00
	lcd_hexb(bcd32[1]);	// Digits of whole volts
	lcd_putch('.');		// Decimal point
	lcd_hexb(bcd32[0]);	// Digits of 1/10 and 1/100 volt
	lcd_putch('V');		// Units
	delay_s1(15);		// Display 15*0.1=1.5 seconds

	while (1) {				// Main loop
		PORTB ^= 0b00000011; // ### diagnostic - execution time

		do_count();			// Count F/4 for 40 ms, result in fp_x_i24
		u24_bcd32();		// BCD result (MHz) is in bcd32[]: 33 22.11 00
		lcd_goto(0x00);		// Top line, first position

		if (bcd32[2] & 0xF0) // Tens of megahertz are non-zero
			lcd_hexnh(bcd32[2]); // Display them
		else
			lcd_putch(' '); // Space instead of leading zero

		lcd_hexnl(bcd32[2]); // Units of megahertz
		lcd_putch('.');		// Decimal point
		lcd_hexb(bcd32[1]); // Hundreds and tens of kilohertz
		lcd_hexnh(bcd32[0]); // Units of kilohertz
		lcd_puts(msg_mhzr); // "MHz R"

//		Start R,X,SWR calculations by measuring v_in, v_50 and v_ld

		a2d_read(ACH_VIN);	// ACH0 - v_in code in fp_x_u16
		fpcalc(FPC_U16TOFP); // Converted to floating point in fp_x
		v_in=fp_x;			// Keep as a floating point value
		fpcalc(FPC_UP);		// Push to fp_y
		fp_x=MIN_VIN;
		fpcalc(FPC_SUB);	// fp_x=v_in-MIN_VIN
		if (fp_x < 0) {		// This doesn't make FP library call
			err_msg(msg_norf);	// Too low RF input
			continue;		// To the start of main loop
		}

		a2d_read(ACH_V50);	// ACH1 - v_50
		fpcalc(FPC_U16TOFP);
		v_50=fp_x;
		fpcalc(FPC_UP);		// Push to fp_y
		fp_x=MIN_V50;
		fpcalc(FPC_SUB);	// fp_x=v_50-MIN_V50
		if (fp_x < 0) {
			err_msg(msg_open);	// Too low RF current
			continue;		// To the start of main loop
		}

		a2d_read(ACH_VLD);	// ACH2 - v_ld
		fpcalc(FPC_U16TOFP);
		v_ld=fp_x;
		fpcalc(FPC_UP);		// Push to fp_y
		fp_x=MIN_VLD;
		fpcalc(FPC_SUB);	// fp_x=v_ld-MIN_VLD
		if (fp_x < 0) {
			err_msg(msg_shrt);	// Too low RF input
			continue;		// To the start of main loop
		}

//		First check for impossible output from network

		fp_y=v_50;			// Accessing fp_y directly
		fp_x=v_ld;
		fpcalc(FPC_ADD);	// fp_x=v_50+v_ld
		fpcalc(FPC_UP);		// fp_y=v_50+v_ld
		fp_temp=fp_y;		// Keep for future use
		fp_x=v_in;
		fpcalc(FPC_SUB);	// fp_y=(v_50+v_ld)-v_in
		if (fp_x < 0)		// That is if v_in > (v_50+v_ld)
			v_in=fp_temp;	// Then assume v_in=(v_50+v_ld)

//		Square all voltages

		fp_x=v_in;
		fpcalc(FPC_UP);		// Both fp_y=v_in and fp_x=v_in
		fpcalc(FPC_MUL);	// fp_x=v_in*v_in
		sq_vin=fp_x;		// Keep result

		fp_x=v_50;
		fpcalc(FPC_UP);
		fpcalc(FPC_MUL);	// fp_x=v_50*v_50
		sq_v50=fp_x;

		fp_x=v_ld;
		fpcalc(FPC_UP);
		fpcalc(FPC_MUL);	// fp_x=v_ld*v_ld
		sq_vld=fp_x;

//		Second check for impossible output from network

		fp_y=sq_v50;		// And sq_vld is already in fp_x
		fpcalc(FPC_ADD);	// fp_x=v_50^2+v_ld^2
		fp_temp=fp_x;		// Keep for future use
		fp_y=sq_vin;		// fp_y=v_in^2
		fpcalc(FPC_SUB);	// fp_x=v_in^2-(v_50^2+v_ld^2)
		if (fp_x < 0)		// That is if v_in^2 < (v_50^2+v_ld^2)
			sq_vin=fp_temp; // Then assume v_in^2=(v_50^2+v_ld^2)

//		Calculate voltage accross R_ld using Equation 1

		fp_y=sq_vin;
		fp_x=fp_temp;		// fp_temp=v_50^2+v_ld^2
		fpcalc(FPC_SUB);	// fp_x=v_in^2-(v_50^2+v_ld^2)
		fpcalc(FPC_UP);		// Push it to fp_y
		fp_x=2.0;
		fpcalc(FPC_DIV);	// fp_x=(v_in^2-v_50^2-v_ld^2)/2
		fpcalc(FPC_UP);		// fp_y=(v_in^2-v_50^2-v_ld^2)/2
		fp_x=v_50;
		fpcalc(FPC_DIV);	// fp_x=(v_in^2-v_50^2-v_ld^2)/(2*v_50)
		sq_vin=fp_x;		// Now sq_vin is v_r

//		Calculate voltage accross X_ld using Equation 2

		fpcalc(FPC_UP);		// Both fp_x=v_r and fp_y=v_r
		fpcalc(FPC_MUL);	// fp_x=v_r^2
		fp_y=sq_vld;		// fp_y=v_ld^2
		fpcalc(FPC_SUB);	// fp_x=v_x^2=v_ld^2-v_r^2
		if (fp_x < 0)		// Just in case
			fp_x=-fp_x;
		fpcalc(FPC_SQRT);	// fp_x=v_x

//		Convert voltages to R and X

		fp_y=50.0;			// And fp_x=v_x
		fpcalc(FPC_MUL);	// fp_x=v_x*50
		fpcalc(FPC_UP);		// Move to fp_y
		fp_x=v_50;
		fpcalc(FPC_DIV);	// fp_x=v_x*50/v_50
		sq_vld=fp_x;		// sq_vld is now X

		fp_x=sq_vin;		// Last time sq_vin was v_r
		fp_y=50.0;
		fpcalc(FPC_MUL);	// fp_x=v_r*50
		fpcalc(FPC_UP);		// Move to fp_y
		fp_x=v_50;
		fpcalc(FPC_DIV);	// fp_x=v_r*50/v_50
		sq_vin=fp_x;		// sq_vin is now R

//		Display R and X

		disp_rx();			// Display R from fp_x
		lcd_goto(0x40);		// Go to line 2
		lcd_putch('X');
		fp_x=sq_vld;		// Last time sq_vld was X
		disp_rx();			// Display X from fp_x

//		Calculate and display SWR. Math moved to a separate routine
//		due to compiler limitations.

		lcd_puts(msg_swr);	// Prepare for SWR display
		fp_x=sq_vld;		// Last time sq_vld was X
		fp_y=sq_vin;		// Last time sq_vin was R
		calc_swr();			// returns SWR in fp_x
		sq_v50=fp_x;		// sq_v50 is now SWR

		fp_y=99.99;			// Upper limit
		fpcalc(FPC_SUB);	// fp_x=99.99-SWR

		if (fp_x < 0) {		// That is if SWR > 99.99
			sq_v50=99.99;	// To use the same display utility
			lcd_putch('>');	// But with "greater than" sign
		} else {
			lcd_putch('=');	// For values up to 99.99
		}

		fp_x=sq_v50;		// Value to display
		fp_y=100.0;			// To get result in hundredth
		fpcalc(FPC_MUL);	// fp_x=SWR*100
		fpcalc(FPC_FPTOS24); // Convert to integer in fp_x_i24
		u24_bcd32();		// BCD result is in bcd32[]: 33 22 11 00
		lcd_hexb(bcd32[1]); // Digits of tens and units
		lcd_putch('.');		// Decimal point
		lcd_hexb(bcd32[0]);	// Digits of tenths and hundredths
		delay_s1(2);		// Wait 0.2s to slow down operation
	}	// Main loop
}


/*
	Local functions - only those can't be moved elsewhere due to free
	compiler limitations on number of arguments and type of return value
*/

/*
	err_msg() - clears second half of first LCD line, then displays
	a message on the second line
*/

void err_msg(const char * s)

{
	char cnt=8;

	lcd_goto(0x09);	// Line 1, beyond frequency value

	do {
		lcd_putch(' ');	// Clear Line 1
	} while (--cnt);

	lcd_goto(0x40);	// Line 2
	lcd_puts(s);	// Error message on Line 2
}


/*
	lcd_puts() - write a string of chars to the LCD
*/

void lcd_puts(const char * s)

{
	LCD_RS = 1;		// Write characters, not commands
	while(*s)
		lcd_write(*s++);
}


/*
	disp_rx() - to be moved into lcd_nmix module
*/

void disp_rx(void)				/* Display R or X */

{
	float fp_temp;

	fp_temp=fp_x;		// Keep fp_x=R for future use
	fp_y=999.0;			// Upper limit
	fpcalc(FPC_SUB);	// fp_x=999.0-R

	if (fp_x < 0) {		// That is if R > 999.0
		fp_temp=999.0;	// To use the same display utility
		lcd_putch('>');	// But with "greater than" sign
	} else {
		lcd_putch('=');	// For values up to 999.0
	}

	fp_x=fp_temp;		// Value to display
	fpcalc(FPC_FPTOS24); // Convert to integer in fp_x_i24
	u24_bcd32();		// BCD result is in bcd32[]: 33 22 11 00
	lcd_hexnl(bcd32[1]); // Digits of thousands of ohm
	lcd_hexb(bcd32[0]);	// Digits of tens and units of ohm
	lcd_putch(LCD_OHM);	// Units
}
