/*
	lcd_nmix.c - LCD and other mixed functions of Ant_Anl project
	Copyright (C) 2005 Vadim Demidov vadim_d .at. sourceforge.net

    Ant_Anl is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published
    by the Free Software Foundation; either version 2 of the License,
    or (at your option) any later version.

    Ant_Anl is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Ant_Anl; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	Revision History:

	17 Apr 05 - Created (VD)
	23 Apr 05 - Added delay_s1() function (VD)
*/

#include "aa_defs.h"	// Definitions common to all Ant_Anl project
#include "glb_defs.h"	// Global functions and data objects

/* Local fuctions */

void ms40(void);		// Precision 40 ms delay
void chk_t0if(void);	// Timer 0 overflow check

#pragma rambank 0
#pragma codepage 0

#define	LCD_STROBE	LCD_EN=1;nop2();nop2();LCD_EN=0

void delay_s1(uns8 s1_cnt)		/* Delay in 0.1 seconds, zero gives 25.6 s */

{
  do {
	delay_ms(100);
  }	while (--s1_cnt);
}


void delay_ms (uns8 ms_cnt)	// Delay in milliseconds, zero argument gives 256 ms

{
  do {
	delay_us(250);
	delay_us(250);
	delay_us(250);
	delay_us(250);
  }	while (--ms_cnt);
}


void delay_us (uns8 us_cnt)	// Delay in microseconds

{	// 3 cycles on function entry. At 16MHZ crystal we have 4 cycles per 1us
	#asm
us_loop:
	NOP					// 1 cycle
	DECFSZ	us_cnt, f	// 1 cycle
	GOTO	us_loop		// 2 cycles
	#endasm
}	// 2 clocks on return


/* write a byte to the LCD in 4 bit mode */

void lcd_write(char c)

{
	PORTB &= 0x0F;
	PORTB |= (c & 0xF0);
	LCD_STROBE;
	PORTB &= 0x0F;
	PORTB |=  (c << 4);
	LCD_STROBE;
	delay_us(40);
}


/*	Clear and home the LCD */

void lcd_clear(void)

{
	LCD_RS = 0;
	lcd_write(0x1);
	delay_ms(2);
}


/* write one character to the LCD */

void lcd_putch(char c)

{
	LCD_RS = 1;	// write characters
	lcd_write(c);
//	PORTB &= 0x0F;
//	PORTB |= (c & 0xF0);
//	LCD_STROBE;
//	PORTB &= 0x0F;
//	PORTB |=  (c << 4);
//	LCD_STROBE;
//	delay_us(40);
}


/*	Go to the specified position */

void lcd_goto(unsigned char pos)

{
	LCD_RS = 0;
	lcd_write(0x80+pos);
}


/* initialise the LCD - put into 4 bit mode */

void lcd_init(void)

{
	LCD_EN=0;	// Must be low after port initialization
	LCD_RS = 0;	// write control bytes
	delay_ms(50);	// power on delay (100ms)
	PORTB &= 0x0F;
	PORTB |= (0x3 << 4);	// attention!
	LCD_STROBE;
	delay_ms(5);	// (100ms)
	LCD_STROBE;
	delay_us(100);	// (200us)
	LCD_STROBE;
	delay_us(100);	// (200us)
//	PORTB &= 0x0F;
//	PORTB |= (0x2 << 4);	// set 4 bit mode
//	LCD_STROBE;
//	delay_us(40);	// (200us)
	lcd_write(0x28);	// 4 bit mode, 1/16 duty, 5x8 font
//	lcd_write(0x08);	// display off
//	lcd_write(0x0F);	// display on, blink curson on
	lcd_write(0x0C);	// display on, blink curson off ?
	lcd_write(0x06);	// entry mode
	lcd_clear();
}


void lcd_hexb(char c) /* Hexadecimal byte output */

{
	char sym;

	sym = (c >> 4);
	if (sym < 10)
		lcd_putch(sym+'0');
	else
		lcd_putch(sym+('A'-10));
	sym = (c & 0x0F);
	if (sym < 10)
		lcd_putch(sym+'0');
	else
		lcd_putch(sym+('A'-10));
}


void lcd_hexnh(char c) /* Hexadecimal high nibble output */

{
	c=c >> 4;
	if (c < 10)
		lcd_putch(c+'0');
	else
		lcd_putch(c+('A'-10));
}


void lcd_hexnl(char c) /* Hexadecimal low nibble output */

{
	c=c & 0x0F;
	if (c < 10)
		lcd_putch(c+'0');
	else
		lcd_putch(c+('A'-10));
}


/*
	do_count() - Count input frequency. Uses the same algorithm as
	Phil's (VK3BHR) "Measure" routine from PIC-based frequency meter.
*/

void do_count(void)

{
	CLAMP=0;	// Although it must be already 0
	T0IF=0;		// Clear any previous overflow
	TMR0=0;		// Clear both timer 0 and its prescaler

	fp_x_i24.low8=0;
//	fp_x_i24.mid8=0;
	fp_x_i24.high8=0;	// Prepare overflow counter

	CLAMP=1;	// Enable counting
	ms40();	// (2+159995)
	CLAMP=0;	// (3) - including two BCF RPx
	nop2();
	nop2();
	nop2();
	chk_t0if();		// Final check, just in case
	fp_x_i24.mid8=TMR0;	// bits [15:8] of total result

	do {
		T0SE=0;		// Toggle prescaler
		nop();
		T0SE=1;
		--fp_x_i24.low8;	// Decrement counter
	} while (fp_x_i24.mid8 == TMR0); // No overflow from prescaler
}


/*
	ms40() - Precision 40 ms delay. Periodically checks for
	Timer 0 overflow by calling chk_t0if(). For 60 MHz input
	frequency the timer can overflow at (60*10^6)/(2^16) or
	915 Hz rate (T=1.09 ms). For 16 MHz crystal this is at least
	once per 4360 instruction cycles. The routine below does it
	once per 22 instruction cycles. If you have modified the code
	(or just used another compiler), carefully check its timing,
	and adjust if necessary. 40 ms is 160000 instruction cycles
	less 5 cycles (2 for CALL and 3 for CLAMP=1), or 159995.

	Total time for N=0x10000-cnt-1 is 22*N+28+A. For N=7271 A=5:
	22*7271+28+5=159995. cnt=0x10000-N-1=58264 (0xE398)
*/

void ms40(void)

{
//	static uns16 cnt;
	uns16 cnt;

	cnt=58264;	// (6), or 4 if one byte is zero

	while (1) {
		chk_t0if();		// (2+8)
		if (++cnt.low8)	// (3)
			goto r01;	// (2-go/1-skip)
		cnt.high8=incsz(cnt.high8); // (3)
			goto r02;	// (2-go/1-skip)
		break;			// (2)
r01:					// 15 from while
		nop2();			// (2)
		nop2();			// (2)
r02:					// 19 from while
		nop();			// (1)
	}					// (2) Full loop 22 inst. cycles

	nop2();				// (2) Goes to "A" delay - out of loop
	nop2();				// (2)
	nop();				// (1) Total A=5
}						// (2) - return


/*
	chk_t0if() - check for Timer 0 interrupt. Must always take
	the same amount of execution time. If you have modified the code
	(or just used another compiler), carefully check its timing
	and account for it in the ms40() accordingly
*/

void chk_t0if(void)

{
	if (!T0IF)		// (1) No overflow yet?
		goto r03;	// (2-go/1-skip) No, go to exit
	T0IF=0;			// (1) Yes, clear flag
	++fp_x_i24.high8; // (3)
	return;			// (2), total 8

r03:				// 3 from start
	nop2();			// (2) Equalize delay
	nop();			// (1)
}					// (2), total 8


/*	Analog to digital converter (ADC or A2D) functions	*/

/*
	a2d_init() - selects ADC mode and turns it on, but doesn't
	start any conversion
*/

void a2d_init(void)

{
	ADCON1=0x80; // select right justify result. A/D port configuration 0
	ADCON0=0x81; // select Fosc/32 for Fosc up to 20 MHz, ADON=1
}


/*
	a2d_read() - reads the specified ADC channel and leaves the result
	in fp_x_u16 ready for conversion to floating point. Similar to ADC
	example from High Tech C compiler
*/

void a2d_read(unsigned char channel)

{
	channel&=0x07;		// truncate channel to 3 bits
	channel<<=3;		// Shift to correct position
	ADCON0&=0xC5;		// clear current channel select
	ADCON0|=(channel);	// apply the new channel select
	delay_us(20);		// For hold capacitor to charge
	GO=1;				// Start conversion on the selected channel
	while(GO)			// Wait for completion
		continue;
	fp_x_u16.high8=ADRESH;	// Two MSBs of the result
	fp_x_u16.low8=ADRESL;	// Lowest 8 bits of the result
}


/*
	uns24_bcd32() - convert a 24-bit unsigned value from fp_x_i24 to
	a 32-bit binary coded decimal (BCD) value and stores it in bcd32
*/

char bcd32[4];			// Global 32-bit BCD variable

void u24_bcd32(void)

{
	char bit_cnt=24;
	char byte_cnt;
	char *chpt;

	bcd32[0]=0;
	bcd32[1]=0;
	bcd32[2]=0;
	bcd32[3]=0;

	do {
		chpt=bcd32;
		byte_cnt=4;

		do {
			*chpt+=0x03;
			if (!(*chpt & 0x08)) // There was < 5 before addition
				*chpt-=0x03;
			*chpt+=0x30;
			if (!(*chpt & 0x80)) // There was < 0x50 before addition
				*chpt-=0x30;
			++chpt;
		} while (-- byte_cnt);

		fp_x_i24.low8=rl(fp_x_i24.low8);
		fp_x_i24.mid8=rl(fp_x_i24.mid8);
		fp_x_i24.high8=rl(fp_x_i24.high8);

		chpt=bcd32;
		*chpt=rl(*chpt);
		++chpt;
		*chpt=rl(*chpt);
		++chpt;
		*chpt=rl(*chpt);
		++chpt;
		*chpt=rl(*chpt);
	} while (-- bit_cnt);
}


void calc_swr(void)		/* Calculates SWR from R and X */

{
	float sq_vld, sq_vin, sq_v50;	// Local variables

	sq_vld=fp_x;	// X was passed in fp_x
	sq_vin=fp_y;	// R was passed in fp_y

	fpcalc(FPC_UP);	// fp_y=X
	fpcalc(FPC_MUL); // fp_x=X^2
	sq_vld=fp_x;	// sq_vld is now X^2

	fp_y=sq_vin;	// Last time sq_vin was R
	fp_x=50.0;
	fpcalc(FPC_ADD); // fp_x=R+50.0
	fpcalc(FPC_UP); // fp_y=R+50.0
	fpcalc(FPC_MUL); // fp_x=(R+50.0)^2
	sq_v50=fp_x;	// sq_v50 is now (R+50.0)^2

	fp_y=sq_vin;	// Last time sq_vin was R
	fp_x=50.0;
	fpcalc(FPC_SUB); // fp_x=R-50.0
	fpcalc(FPC_UP); // fp_y=R-50.0
	fpcalc(FPC_MUL); // fp_x=(R-50.0)^2
	fp_y=sq_vld;	// Last time sq_vld was X^2
	fpcalc(FPC_ADD); // fp_x=(R-50.0)^2+X^2
	fpcalc(FPC_SQRT); // fp_x=B
	sq_vin=fp_x;	// sq_vin is now B

	fp_y=sq_v50;	// Last time sq_v50 was (R+50.0)^2
	fp_x=sq_vld;	// Last time sq_vld was X^2
	fpcalc(FPC_ADD); // fp_x=(R+50.0)^2+X^2
	fpcalc(FPC_SQRT); // fp_x=A
	sq_vld=fp_x;	// sq_vld is now A

	fp_y=sq_vin;	// Last time sq_vin was B
	fpcalc(FPC_SUB); // fp_x=B-A
	sq_v50=-fp_x;	// sq_v50 is now A-B
	fp_x=sq_vld;	// fp_x=A, and fp_y is still B
	fpcalc(FPC_ADD); // fp_x=A+B
	fpcalc(FPC_UP);	// fp_y=A+B
	fp_x=sq_v50;	// Last time sq_v50 was A-B
	fpcalc(FPC_DIV); // fp_x=(A+B)/(A-B) or SWR
}
